@@ -21,20 +21,19 @@ from elementpath.datatypes import MONTH_DAYS, MONTH_DAYS_LEAP, days_from_common_
     months2days, DateTime, DateTime10, Date, Date10, Time, Timezone, Duration, \
     DayTimeDuration, YearMonthDuration, UntypedAtomic, GregorianYear, GregorianYear10, \
     GregorianYearMonth, GregorianYearMonth10, GregorianMonthDay, GregorianMonth, \
-    GregorianDay, AbstractDateTime, OrderedDateTime, NumericTypeProxy, ArithmeticTypeProxy, \
-    decimal_validator, integer_validator, base64_binary_validator, hex_binary_validator, \
-    ncname_validator, XSD_BUILTIN_TYPES, XsdBuiltin, is_id, is_idrefs
+    GregorianDay, AbstractDateTime, OrderedDateTime, NumericProxy, ArithmeticProxy, \
+    Id, is_idrefs
 
 
-class DatatypesHelpersTest(unittest.TestCase):
+class StringTypesTest(unittest.TestCase):
 
     def test_is_id_function(self):
-        self.assertTrue(is_id(ElementTree.XML('<A>xyz</A>').text))
-        self.assertFalse(is_id(ElementTree.XML('<A>xyz abc</A>').text))
-        self.assertFalse(is_id(ElementTree.XML('<A>12345</A>').text))
-        self.assertTrue(is_id('alpha'))
-        self.assertFalse(is_id('alpha beta'))
-        self.assertFalse(is_id('12345'))
+        self.assertTrue(Id.is_valid(ElementTree.XML('<A>xyz</A>').text))
+        self.assertFalse(Id.is_valid(ElementTree.XML('<A>xyz abc</A>').text))
+        self.assertFalse(Id.is_valid(ElementTree.XML('<A>12345</A>').text))
+        self.assertTrue(Id.is_valid('alpha'))
+        self.assertFalse(Id.is_valid('alpha beta'))
+        self.assertFalse(Id.is_valid('12345'))
 
     def test_node_is_idref_function(self):
         self.assertTrue(is_idrefs(ElementTree.XML('<A>xyz</A>').text))
@@ -53,6 +52,8 @@ class UntypedAtomicTest(unittest.TestCase):
         self.assertEqual(UntypedAtomic('alpha').value, 'alpha')
         self.assertEqual(UntypedAtomic(b'beta').value, 'beta')
         self.assertEqual(UntypedAtomic(True).value, 'true')
+        self.assertEqual(UntypedAtomic(UntypedAtomic(2)).value, '2')
+        self.assertEqual(UntypedAtomic(Date.fromstring('2000-02-01')).value, '2000-02-01')
 
         with self.assertRaises(TypeError) as err:
             UntypedAtomic(None)
@@ -79,7 +80,7 @@ class UntypedAtomicTest(unittest.TestCase):
         self.assertTrue(bool(True) == UntypedAtomic(1))
         with self.assertRaises(ValueError) as ctx:
             _ = bool(True) == UntypedAtomic(10)
-        self.assertEqual(str(ctx.exception), "'10' cannot be cast to boolean")
+        self.assertEqual(str(ctx.exception), "'10' cannot be cast to xs:boolean")
 
         self.assertFalse(-10.9 == UntypedAtomic(-10))
         self.assertFalse(UntypedAtomic(-10) == -11)
@@ -172,7 +173,7 @@ class DateTimeTypesTest(unittest.TestCase):
 
         with self.assertRaises(TypeError) as err:
             DateTime(year=-1999.0, month=1, day=1)
-        self.assertIn("wrong type <class 'float'> for year", str(err.exception))
+        self.assertIn("invalid type <class 'float'> for year", str(err.exception))
 
     def test_datetime_fromstring(self):
         dt = DateTime.fromstring('2000-10-07T00:00:00')
@@ -197,7 +198,11 @@ class DateTimeTypesTest(unittest.TestCase):
         self.assertIn("2nd argument has an invalid type <class 'str'>", str(err.exception))
 
         with self.assertRaises(ValueError) as err:
-            DateTime.fromstring('00-10-07')
+            DateTime.fromstring('2000-10-07')
+        self.assertIn("Invalid datetime string", str(err.exception))
+
+        with self.assertRaises(ValueError) as err:
+            DateTime.fromstring('00-10-07T00:00:00')
         self.assertIn("Invalid datetime string", str(err.exception))
 
         with self.assertRaises(ValueError) as err:
@@ -209,6 +214,12 @@ class DateTimeTypesTest(unittest.TestCase):
         self.assertIsInstance(Date.fromstring('-2000-10-07'), Date)
         self.assertIsInstance(Date.fromstring('0000-02-29'), Date)
 
+        with self.assertRaises(ValueError) as ctx:
+            Date.fromstring('01000-02-29')
+
+        self.assertIn("when year exceeds 4 digits leading zeroes are not allowed",
+                      str(ctx.exception))
+
         dt = Date.fromstring("-0003-01-01")
         self.assertEqual(dt._year, -4)
         self.assertEqual(dt._dt.year, 6)
@@ -238,7 +249,7 @@ class DateTimeTypesTest(unittest.TestCase):
         self.assertEqual(DateTime10(-1, 10, 7).iso_year, '-0001')
 
     def test_datetime_repr(self):
-        dt = DateTime.fromstring('2000-10-07')
+        dt = DateTime.fromstring('2000-10-07T00:00:00')
         self.assertEqual(repr(dt), "DateTime(2000, 10, 7, 0, 0, 0)")
         self.assertEqual(str(dt), '2000-10-07T00:00:00')
 
@@ -259,7 +270,7 @@ class DateTimeTypesTest(unittest.TestCase):
 
         dt = DateTime(2001, 1, 1, microsecond=10)
         self.assertEqual(repr(dt), 'DateTime(2001, 1, 1, 0, 0, 0.000010)')
-        self.assertEqual(str(dt), '2001-01-01T00:00:00.000010')
+        self.assertEqual(str(dt), '2001-01-01T00:00:00.00001')
 
     def test_24_hour_datetime(self):
         dt = DateTime.fromstring('0000-09-19T24:00:00Z')
@@ -364,13 +375,13 @@ class DateTimeTypesTest(unittest.TestCase):
         self.assertTrue(mkdt("2005-04-04T24:00:00", tz) == mkdt("2005-04-05T00:00:00", tz))
         self.assertFalse(mkdt("2005-04-04T24:00:00", tz) != mkdt("2005-04-05T00:00:00", tz))
 
-        self.assertTrue(DateTime.fromstring("-1000-01-01") == DateTime.fromstring("-1000-01-01"))
-        self.assertTrue(DateTime.fromstring("-10000-01-01") == DateTime.fromstring("-10000-01-01"))
-        self.assertFalse(DateTime.fromstring("20000-01-01") != DateTime.fromstring("20000-01-01"))
-        self.assertFalse(DateTime.fromstring("-10000-01-02") == DateTime.fromstring("-10000-01-01"))
+        self.assertTrue(Date.fromstring("-1000-01-01") == Date.fromstring("-1000-01-01"))
+        self.assertTrue(Date.fromstring("-10000-01-01") == Date.fromstring("-10000-01-01"))
+        self.assertFalse(Date.fromstring("20000-01-01") != Date.fromstring("20000-01-01"))
+        self.assertFalse(Date.fromstring("-10000-01-02") == Date.fromstring("-10000-01-01"))
 
-        self.assertFalse(DateTime.fromstring("-10000-01-02") == (1, 2, 3))  # Wrong type
-        self.assertTrue(DateTime.fromstring("-10000-01-02") != (1, 2, 3))  # Wrong type
+        self.assertFalse(Date.fromstring("-10000-01-02") == (1, 2, 3))  # Wrong type
+        self.assertTrue(Date.fromstring("-10000-01-02") != (1, 2, 3))  # Wrong type
 
     def test_lt_operator(self):
         mkdt = DateTime.fromstring
@@ -882,12 +893,12 @@ class DurationTypesTest(unittest.TestCase):
 
         with self.assertRaises(TypeError) as err:
             _ = year_month_duration('P2Y') + daytime_duration('P1D')
-        self.assertIn("wrong type <class 'elementpath.datatypes.DayTimeDuration'",
+        self.assertIn("cannot add <class 'elementpath.datatypes.DayTimeDuration'",
                       str(err.exception))
 
         with self.assertRaises(TypeError) as err:
             _ = daytime_duration('P1D') + year_month_duration('P2Y')
-        self.assertIn("wrong type <class 'elementpath.datatypes.YearMonthDuration'",
+        self.assertIn("cannot add <class 'elementpath.datatypes.YearMonthDuration'",
                       str(err.exception))
 
     def test_sub_operator(self):
@@ -902,12 +913,12 @@ class DurationTypesTest(unittest.TestCase):
 
         with self.assertRaises(TypeError) as err:
             _ = year_month_duration('P2Y') - daytime_duration('P1D')
-        self.assertIn("wrong type <class 'elementpath.datatypes.DayTimeDuration'",
+        self.assertIn("cannot subtract <class 'elementpath.datatypes.DayTimeDuration'",
                       str(err.exception))
 
         with self.assertRaises(TypeError) as err:
             _ = daytime_duration('P1D') - year_month_duration('P2Y')
-        self.assertIn("wrong type <class 'elementpath.datatypes.YearMonthDuration'",
+        self.assertIn("cannot subtract <class 'elementpath.datatypes.YearMonthDuration'",
                       str(err.exception))
 
     def test_mul_operator(self):
@@ -918,13 +929,13 @@ class DurationTypesTest(unittest.TestCase):
 
         with self.assertRaises(TypeError) as err:
             _ = daytime_duration('P1D') * '2'
-        self.assertEqual("wrong type <class 'str'> for operand '2'", str(err.exception))
+        self.assertIn("cannot multiply", str(err.exception))
 
         self.assertEqual(year_month_duration('P1Y2M') * 2, YearMonthDuration(months=14 * 2))
 
         with self.assertRaises(TypeError) as err:
             _ = year_month_duration('P1Y2M') * '2'
-        self.assertEqual("wrong type <class 'str'> for operand '2'", str(err.exception))
+        self.assertIn("cannot multiply", str(err.exception))
 
     def test_div_operator(self):
         daytime_duration = DayTimeDuration.fromstring
@@ -934,13 +945,13 @@ class DurationTypesTest(unittest.TestCase):
 
         with self.assertRaises(TypeError) as err:
             _ = daytime_duration('P2D') / '2'
-        self.assertEqual("wrong type <class 'str'> for operand '2'", str(err.exception))
+        self.assertIn("cannot divide", str(err.exception))
 
         self.assertEqual(year_month_duration('P1Y2M') / 2, YearMonthDuration(months=14 / 2))
 
         with self.assertRaises(TypeError) as err:
             _ = year_month_duration('P1Y2M') / '2'
-        self.assertEqual("wrong type <class 'str'> for operand '2'", str(err.exception))
+        self.assertIn("cannot divide", str(err.exception))
 
     def test_hashing(self):
         self.assertIsInstance(hash(Duration(16)), int)
@@ -1032,68 +1043,36 @@ class TimezoneTypeTest(unittest.TestCase):
 class TypeProxiesTest(unittest.TestCase):
 
     def test_instance_check(self):
-        self.assertIsInstance(10, NumericTypeProxy)
-        self.assertIsInstance(17.8, NumericTypeProxy)
-        self.assertIsInstance(Decimal('18.12'), NumericTypeProxy)
-        self.assertNotIsInstance(True, NumericTypeProxy)
-        self.assertNotIsInstance(Duration.fromstring('P1Y'), NumericTypeProxy)
+        self.assertIsInstance(10, NumericProxy)
+        self.assertIsInstance(17.8, NumericProxy)
+        self.assertIsInstance(Decimal('18.12'), NumericProxy)
+        self.assertNotIsInstance(True, NumericProxy)
+        self.assertNotIsInstance(Duration.fromstring('P1Y'), NumericProxy)
 
-        self.assertIsInstance(10, ArithmeticTypeProxy)
+        self.assertIsInstance(10, ArithmeticProxy)
 
     def test_subclass_check(self):
-        self.assertFalse(issubclass(bool, NumericTypeProxy))
-        self.assertFalse(issubclass(str, NumericTypeProxy))
-        self.assertTrue(issubclass(int, NumericTypeProxy))
-        self.assertTrue(issubclass(float, NumericTypeProxy))
-        self.assertTrue(issubclass(Decimal, NumericTypeProxy))
-        self.assertFalse(issubclass(DateTime10, NumericTypeProxy))
+        self.assertFalse(issubclass(bool, NumericProxy))
+        self.assertFalse(issubclass(str, NumericProxy))
+        self.assertTrue(issubclass(int, NumericProxy))
+        self.assertTrue(issubclass(float, NumericProxy))
+        self.assertTrue(issubclass(Decimal, NumericProxy))
+        self.assertFalse(issubclass(DateTime10, NumericProxy))
 
-        self.assertFalse(issubclass(bool, ArithmeticTypeProxy))
-        self.assertFalse(issubclass(str, ArithmeticTypeProxy))
-        self.assertTrue(issubclass(int, ArithmeticTypeProxy))
-        self.assertTrue(issubclass(float, ArithmeticTypeProxy))
-        self.assertTrue(issubclass(Decimal, ArithmeticTypeProxy))
+        self.assertFalse(issubclass(bool, ArithmeticProxy))
+        self.assertFalse(issubclass(str, ArithmeticProxy))
+        self.assertTrue(issubclass(int, ArithmeticProxy))
+        self.assertTrue(issubclass(float, ArithmeticProxy))
+        self.assertTrue(issubclass(Decimal, ArithmeticProxy))
 
     # noinspection PyArgumentList
     def test_instance_build(self):
-        self.assertEqual(NumericTypeProxy(), 0.0)
-        self.assertEqual(NumericTypeProxy(9), 9.0)
-        self.assertEqual(NumericTypeProxy('49'), 49.0)
-        self.assertEqual(ArithmeticTypeProxy(), 0.0)
-        self.assertEqual(ArithmeticTypeProxy(8.0), 8.0)
-        self.assertEqual(ArithmeticTypeProxy('81.0'), 81.0)
-
-
-class XsdBuiltinTypesTest(unittest.TestCase):
-
-    def test_decimal_validator(self):
-        self.assertTrue(decimal_validator(Decimal('10.9')))
-        self.assertFalse(decimal_validator(10.9))
-
-    def test_integer_validator(self):
-        self.assertTrue(integer_validator(10))
-        self.assertFalse(integer_validator(10.0))
-        self.assertFalse(integer_validator(Decimal('10.0')))
-
-    def test_base64_binary_validator(self):
-        self.assertTrue(base64_binary_validator('YWxwaGE='))
-        self.assertTrue(base64_binary_validator(b'YWxwaGE='))
-        self.assertFalse(base64_binary_validator('YWxwaGE!='))
-        self.assertFalse(base64_binary_validator('XYZ'))
-        self.assertFalse(base64_binary_validator('=59A'))
-
-    def test_hex_binary_validator(self):
-        self.assertTrue(hex_binary_validator('F3'))
-        self.assertFalse(hex_binary_validator('G3'))
-
-    def test_ncname_validator(self):
-        self.assertTrue(ncname_validator('myElement'))
-        self.assertFalse(ncname_validator('tns:myElement'))
-
-    def test_types(self):
-        self.assertEqual(len(XSD_BUILTIN_TYPES), 47)
-        for name in XSD_BUILTIN_TYPES:
-            self.assertIsInstance(XSD_BUILTIN_TYPES[name], XsdBuiltin)
+        self.assertEqual(NumericProxy(), 0.0)
+        self.assertEqual(NumericProxy(9), 9.0)
+        self.assertEqual(NumericProxy('49'), 49.0)
+        self.assertEqual(ArithmeticProxy(), 0.0)
+        self.assertEqual(ArithmeticProxy(8.0), 8.0)
+        self.assertEqual(ArithmeticProxy('81.0'), 81.0)
 
 
 if __name__ == '__main__':
