@@ -9,6 +9,7 @@
 # @author Davide Brunato <brunato@sissa.it>
 #
 import unittest
+from unittest.mock import patch
 import datetime
 import platform
 from decimal import Decimal
@@ -19,10 +20,10 @@ except ImportError:
     lxml_etree = None
 
 from elementpath import XPathContext, AttributeNode, TypedAttribute
-from elementpath.datatypes import Timezone, DateTime, GregorianYear10, \
-    Duration, YearMonthDuration, DayTimeDuration, Date10, Time, \
-    XPathGregorianDay, XPathGregorianMonth, XPathGregorianMonthDay, \
-    XPathGregorianYear, XPathGregorianYearMonth, UntypedAtomic
+from elementpath.datatypes import Timezone, DateTime, GregorianDay, \
+    GregorianMonth, GregorianMonthDay, GregorianYear10, GregorianYearMonth10, \
+    Duration, YearMonthDuration, DayTimeDuration, Date10, Time, QName, UntypedAtomic
+from elementpath.namespaces import XSD_NAMESPACE
 
 try:
     from tests import xpath_test_class
@@ -33,7 +34,8 @@ except ImportError:
 class XPath2ConstructorsTest(xpath_test_class.XPathTestCase):
 
     def test_string_constructor(self):
-        self.check_value("xs:string(5.0)", '5.0')
+        self.check_value("xs:string(5.0)", '5')
+        self.check_value("xs:string(5.2)", '5.2')
         self.check_value('xs:string(" hello  ")', ' hello  ')
         self.check_value('xs:string("\thello \n")', '\thello \n')
         self.check_value('xs:string(())', [])
@@ -132,12 +134,12 @@ class XPath2ConstructorsTest(xpath_test_class.XPathTestCase):
         self.check_value('xs:ENTITY(xs:untypedAtomic("xyz"))', 'xyz')
 
     def test_qname_constructor(self):
+        qname = QName(XSD_NAMESPACE, 'xs:element')
         self.check_value('xs:QName(())', [])
-        self.check_value('xs:QName("xs:element")', 'xs:element')
-        self.check_value('xs:QName(xs:untypedAtomic("xs:element"))', 'xs:element')
-        self.assertRaises(KeyError, self.parser.parse, 'xs:QName("xsd:element")')
-        self.wrong_type('xs:QName(5)', 'FORG0006', "the argument has an invalid type")
-        self.wrong_value('xs:QName("1")', 'FOCA0002', "the argument must be an xs:QName")
+        self.check_value('xs:QName("xs:element")', qname)
+        self.check_value('xs:QName(xs:untypedAtomic("xs:element"))', qname)
+        self.wrong_type('xs:QName(5)', 'XPTY0004', "the argument has an invalid type")
+        self.wrong_value('xs:QName("1")', 'FORG0001', "invalid value")
 
     def test_any_uri_constructor(self):
         self.check_value('xs:anyURI("")', '')
@@ -157,7 +159,7 @@ class XPath2ConstructorsTest(xpath_test_class.XPathTestCase):
         if platform.python_version_tuple() >= ('3', '6') and \
                 platform.python_implementation() != 'PyPy':
             self.wrong_value('xs:anyURI("https://example.com:65536")',
-                             'FOCA0002', 'Port out of range 0-65535')
+                             'FORG0001', 'Port out of range 0-65535')
 
         root = self.etree.XML('<root a=" https://example.com "/>')
         context = XPathContext(root)
@@ -169,14 +171,19 @@ class XPath2ConstructorsTest(xpath_test_class.XPathTestCase):
         self.check_value('xs:boolean(0)', False)
         self.check_value('xs:boolean(xs:boolean(0))', False)
         self.check_value('xs:boolean(xs:untypedAtomic(0))', False)
-        self.wrong_type('xs:boolean(xs:hexBinary("FF"))', 'FORG0006', "<class 'bytes'>")
-        self.wrong_value('xs:boolean("2")', 'FOCA0002', "not a boolean value")
+        self.wrong_type('xs:boolean(xs:hexBinary("FF"))', 'XPTY0004', "HexBinary")
+        self.wrong_value('xs:boolean("2")', 'FORG0001', "invalid value")
 
     def test_integer_constructors(self):
         self.wrong_value('xs:integer("hello")', 'FORG0001')
         self.check_value('xs:integer("19")', 19)
         self.check_value('xs:integer(xs:untypedAtomic("19"))', 19)
         self.check_value("xs:integer('-5')", -5)
+        self.wrong_value("xs:integer('INF')", 'FORG0001')
+        self.check_value("xs:integer('inf')", ValueError)
+        self.wrong_value("xs:integer('NaN')", 'FORG0001')
+        self.wrong_value("xs:integer(xs:float('-INF'))", 'FOCA0002')
+        self.check_value("xs:integer(xs:double('NaN'))", ValueError)
 
         root = self.etree.XML('<root a="19"/>')
         context = XPathContext(root)
@@ -203,9 +210,9 @@ class XPath2ConstructorsTest(xpath_test_class.XPathTestCase):
         self.wrong_value('xs:long("340282366920938463463374607431768211456")')
         self.check_value('xs:long("-20")', -20)
         self.wrong_value('xs:int("-20 91")')
-        self.wrong_value('xs:int("9223372036854775808")')
-        self.check_value('xs:int("-9223372036854775808")', -2**63)
-        self.check_value('xs:int("4611686018427387904")', 2**62)
+        self.wrong_value('xs:int("2147483648")')
+        self.check_value('xs:int("2147483647")', 2**31 - 1)
+        self.check_value('xs:int("-2147483648")', -2**31)
         self.wrong_value('xs:short("40000")')
         self.check_value('xs:short("9999")', 9999)
         self.check_value('xs:short(-9999)', -9999)
@@ -217,8 +224,8 @@ class XPath2ConstructorsTest(xpath_test_class.XPathTestCase):
 
         self.wrong_value('xs:unsignedLong("-10")')
         self.check_value('xs:unsignedLong("3")', 3)
-        self.wrong_value('xs:unsignedInt("-9223372036854775808")')
-        self.check_value('xs:unsignedInt("9223372036854775808")', 2**63)
+        self.wrong_value('xs:unsignedInt("-4294967296")')
+        self.check_value('xs:unsignedInt("4294967295")', 2**32 - 1)
         self.wrong_value('xs:unsignedShort("-1")')
         self.check_value('xs:unsignedShort("0")', 0)
         self.wrong_value('xs:unsignedByte(-128)')
@@ -248,10 +255,14 @@ class XPath2ConstructorsTest(xpath_test_class.XPathTestCase):
 
     def test_float_constructor(self):
         self.wrong_value('xs:float("..")')
+        self.wrong_value('xs:float("ab")')
         self.check_value('xs:float(25.05)', 25.05)
         self.check_value('xs:float(xs:untypedAtomic(25.05))', 25.05)
         self.check_value('xs:float(-0.00001)', -0.00001)
         self.check_value('xs:float(0.00001)', float)
+        self.check_value('xs:float("INF")', float('inf'))
+        self.check_value('xs:float("-INF")', float('-inf'))
+        self.wrong_value('xs:float("inf")')
 
         root = self.etree.XML('<root a="10.3"/>')
         context = XPathContext(root, item=AttributeNode('a', 10.3))
@@ -333,8 +344,10 @@ class XPath2ConstructorsTest(xpath_test_class.XPathTestCase):
         self.check_value('xs:date(.)', ValueError, context=context)
         context.item = AttributeNode('a', True)
         self.check_value('xs:date(.)', ValueError, context=context)
-        context.item = TypedAttribute(AttributeNode('a', 'true'), True)
-        self.check_value('xs:date(.)', TypeError, context=context)
+
+        with patch.multiple(self.dummy_type, is_simple=lambda x: True):
+            context.item = TypedAttribute(AttributeNode('a', 'true'), self.dummy_type, True)
+            self.check_value('xs:date(.)', TypeError, context=context)
 
         root = self.etree.XML("<root>2017-10-02</root>")
         context = XPathContext(root)
@@ -362,10 +375,10 @@ class XPath2ConstructorsTest(xpath_test_class.XPathTestCase):
 
         root = self.etree.XML('<root a="---08"/>')
         context = XPathContext(root)
-        self.check_value('xs:gDay(@a)', XPathGregorianDay(8), context=context)
+        self.check_value('xs:gDay(@a)', GregorianDay(8), context=context)
 
-        context.item = XPathGregorianDay(10)
-        self.check_value('xs:gDay(.)', XPathGregorianDay(10), context=context)
+        context.item = GregorianDay(10)
+        self.check_value('xs:gDay(.)', GregorianDay(10), context=context)
 
     def test_gregorian_month_constructor(self):
         self.check_value('xs:gMonth("--09")', datetime.datetime(2000, 9, 1))
@@ -377,10 +390,10 @@ class XPath2ConstructorsTest(xpath_test_class.XPathTestCase):
 
         root = self.etree.XML('<root a="--11"/>')
         context = XPathContext(root)
-        self.check_value('xs:gMonth(@a)', XPathGregorianMonth(11), context=context)
+        self.check_value('xs:gMonth(@a)', GregorianMonth(11), context=context)
 
-        context.item = XPathGregorianMonth(1)
-        self.check_value('xs:gMonth(.)', XPathGregorianMonth(1), context=context)
+        context.item = GregorianMonth(1)
+        self.check_value('xs:gMonth(.)', GregorianMonth(1), context=context)
 
     def test_gregorian_month_day_constructor(self):
         tz0 = None
@@ -398,10 +411,10 @@ class XPath2ConstructorsTest(xpath_test_class.XPathTestCase):
 
         root = self.etree.XML('<root a="--05-20"/>')
         context = XPathContext(root)
-        self.check_value('xs:gMonthDay(@a)', XPathGregorianMonthDay(5, 20), context=context)
+        self.check_value('xs:gMonthDay(@a)', GregorianMonthDay(5, 20), context=context)
 
-        context.item = XPathGregorianMonthDay(1, 15)
-        self.check_value('xs:gMonthDay(.)', XPathGregorianMonthDay(1, 15), context=context)
+        context.item = GregorianMonthDay(1, 15)
+        self.check_value('xs:gMonthDay(.)', GregorianMonthDay(1, 15), context=context)
 
     def test_gregorian_year_constructor(self):
         self.check_value('xs:gYear("2004")', datetime.datetime(2004, 1, 1))
@@ -415,10 +428,10 @@ class XPath2ConstructorsTest(xpath_test_class.XPathTestCase):
 
         root = self.etree.XML('<root a="1999"/>')
         context = XPathContext(root)
-        self.check_value('xs:gYear(@a)', XPathGregorianYear(1999), context=context)
+        self.check_value('xs:gYear(@a)', GregorianYear10(1999), context=context)
 
-        context.item = XPathGregorianYear(1492)
-        self.check_value('xs:gYear(.)', XPathGregorianYear(1492), context=context)
+        context.item = GregorianYear10(1492)
+        self.check_value('xs:gYear(.)', GregorianYear10(1492), context=context)
 
     def test_gregorian_year_month_constructor(self):
         self.check_value('xs:gYearMonth("2004-02")',
@@ -430,10 +443,10 @@ class XPath2ConstructorsTest(xpath_test_class.XPathTestCase):
 
         root = self.etree.XML('<root a="1900-01"/>')
         context = XPathContext(root)
-        self.check_value('xs:gYearMonth(@a)', XPathGregorianYearMonth(1900, 1), context=context)
+        self.check_value('xs:gYearMonth(@a)', GregorianYearMonth10(1900, 1), context=context)
 
-        context.item = XPathGregorianYearMonth(1300, 10)
-        self.check_value('xs:gYearMonth(.)', XPathGregorianYearMonth(1300, 10), context=context)
+        context.item = GregorianYearMonth10(1300, 10)
+        self.check_value('xs:gYearMonth(.)', GregorianYearMonth10(1300, 10), context=context)
 
     def test_duration_constructor(self):
         self.check_value('xs:duration("P3Y5M1D")', (41, 86400))
@@ -441,7 +454,7 @@ class XPath2ConstructorsTest(xpath_test_class.XPathTestCase):
         self.check_value('xs:duration("P3Y5M1DT1H")', (41, 90000))
         self.check_value('xs:duration("P3Y5M1DT1H3M2.01S")', (41, Decimal('90182.01')))
         self.wrong_value('xs:duration("P3Y5M1X")')
-        self.assertRaises(TypeError, self.parser.parse, 'xs:duration(1)')
+        self.assertRaises(ValueError, self.parser.parse, 'xs:duration(1)')
 
         root = self.etree.XML('<root a="P1Y5M"/>')
         context = XPathContext(root)
@@ -492,56 +505,49 @@ class XPath2ConstructorsTest(xpath_test_class.XPathTestCase):
 
     def test_hex_binary_constructor(self):
         self.check_value('xs:hexBinary(())', [])
-        self.check_value('xs:hexBinary("84")', b'3834')
-        self.check_value('xs:hexBinary(xs:hexBinary("84"))', b'3834')
+        self.check_value('xs:hexBinary("84")', b'84')
+        self.check_value('xs:hexBinary(xs:hexBinary("84"))', b'84')
         self.wrong_type('xs:hexBinary(12)')
 
         root = self.etree.XML('<root a="84"/>')
         context = XPathContext(root)
-        self.check_value('xs:hexBinary(@a)', b'3834', context=context)
+        self.check_value('xs:hexBinary(@a)', b'84', context=context)
 
         context.item = UntypedAtomic('84')
-        self.check_value('xs:hexBinary(.)', b'3834', context=context)
+        self.check_value('xs:hexBinary(.)', b'84', context=context)
 
         context.item = '84'
-        self.check_value('xs:hexBinary(.)', b'3834', context=context)
+        self.check_value('xs:hexBinary(.)', b'84', context=context)
 
         context.item = b'84'
         self.check_value('xs:hexBinary(.)', b'84', context=context)
 
         context.item = b'XY'
-        self.check_value('xs:hexBinary(.)', b'5859', context=context)
+        self.check_value('xs:hexBinary(.)', ValueError, context=context)
 
-        context.item = b'ODQ='
-        if platform.python_implementation() != 'PyPy':
-            # Skip for PyPy because PyPy 7.1.1-beta0 has a bug on codecs.decode()
-            self.check_value('xs:hexBinary(.)', b'3834', context=context)
+        context.item = b'F859'
+        self.check_value('xs:hexBinary(.)', b'F859', context=context)
 
     def test_base64_binary_constructor(self):
         self.check_value('xs:base64Binary(())', [])
-        self.check_value('xs:base64Binary("84")', b'ODQ=\n')
-        self.check_value('xs:base64Binary(xs:base64Binary("84"))', b'ODQ=\n')
-        self.check_value('xs:base64Binary("abcefghi")', b'YWJjZWZnaGk=\n')
+        self.check_value('xs:base64Binary("ODQ=")', b'ODQ=')
+        self.check_value('xs:base64Binary(xs:base64Binary("ODQ="))', b'ODQ=')
+        self.check_value('xs:base64Binary("YWJjZWZnaGk=")', b'YWJjZWZnaGk=')
         self.wrong_type('xs:base64Binary(1e2)')
         self.wrong_type('xs:base64Binary(1.1)')
 
-        root = self.etree.XML('<root a="abcefghi"/>')
+        root = self.etree.XML('<root a="YWJjZWZnaGk="/>')
         context = XPathContext(root)
-        self.check_value('xs:base64Binary(@a)', b'YWJjZWZnaGk=\n', context=context)
+        self.check_value('xs:base64Binary(@a)', b'YWJjZWZnaGk=', context=context)
 
-        context.item = UntypedAtomic('abcefghi')
-        self.check_value('xs:base64Binary(.)', b'YWJjZWZnaGk=\n', context=context)
-
-        context.item = b'FF'
-        if platform.python_implementation() != 'PyPy':
-            # Skip for PyPy because PyPy 7.1.1-beta0 has a bug on codecs.decode()
-            self.check_value('xs:base64Binary(.)', b'/w==\n', context=context)
+        context.item = UntypedAtomic('YWJjZWZnaGk=')
+        self.check_value('xs:base64Binary(.)', b'YWJjZWZnaGk=', context=context)
 
         context.item = b'abcefghi'  # Don't change, it can be an encoded value.
         self.check_value('xs:base64Binary(.)', b'abcefghi', context=context)
 
-        context.item = b'abcefghij'
-        self.check_value('xs:base64Binary(.)', b'YWJjZWZnaGlq\n', context=context)
+        context.item = b'YWJjZWZnaGlq'
+        self.check_value('xs:base64Binary(.)', b'YWJjZWZnaGlq', context=context)
 
     def test_untyped_atomic_constructor(self):
         self.check_value('xs:untypedAtomic(())', [])
