@@ -37,8 +37,9 @@ else:
     xmlschema.XMLSchema.meta_schema.build()
 
 from elementpath import *
-from elementpath.datatypes import DateTime, Date, Time, Timezone, \
-    DayTimeDuration, YearMonthDuration, UntypedAtomic
+from elementpath.datatypes import xsd10_atomic_types, xsd11_atomic_types, DateTime, \
+    Date, Time, Timezone, DayTimeDuration, YearMonthDuration, UntypedAtomic, QName
+from elementpath.xpath_nodes import node_kind
 
 try:
     from tests import test_xpath1_parser
@@ -46,10 +47,50 @@ except ImportError:
     import test_xpath1_parser
 
 
+def get_sequence_type(value, xsd_version='1.0'):
+    """
+    Infers the sequence type from a value.
+    """
+    if value is None or value == []:
+        return 'empty-sequence()'
+    elif isinstance(value, list):
+        if value[0] is not None and not isinstance(value[0], list):
+            sequence_type = get_sequence_type(value[0], xsd_version)
+            if all(get_sequence_type(x, xsd_version) == sequence_type for x in value[1:]):
+                return '{}+'.format(sequence_type)
+            else:
+                return 'node()+'
+    else:
+        value_kind = node_kind(value)
+        if value_kind is not None:
+            return '{}()'.format(value_kind)
+        elif isinstance(value, UntypedAtomic):
+            return 'xs:untypedAtomic'
+
+        if QName.is_valid(value) and ':' in str(value):
+            return 'xs:QName'
+
+        if xsd_version == '1.0':
+            atomic_types = xsd10_atomic_types
+        else:
+            atomic_types = xsd11_atomic_types
+            if atomic_types['dateTimeStamp'].is_valid(value):
+                return 'xs:dateTimeStamp'
+
+        for type_name in ['string', 'boolean', 'decimal', 'float', 'double',
+                          'date', 'dateTime', 'gDay', 'gMonth', 'gMonthDay', 'anyURI',
+                          'gYear', 'gYearMonth', 'time', 'duration', 'dayTimeDuration',
+                          'yearMonthDuration', 'base64Binary', 'hexBinary']:
+            if atomic_types[type_name].is_valid(value):
+                return 'xs:%s' % type_name
+
+    raise ValueError("Inconsistent sequence type for {!r}".format(value))
+
+
 class XPath2ParserTest(test_xpath1_parser.XPath1ParserTest):
 
     def setUp(self):
-        self.parser = XPath2Parser(namespaces=self.namespaces, variables=self.variables)
+        self.parser = XPath2Parser(namespaces=self.namespaces)
 
         # Make sure the tests are repeatable.
         env_vars_to_tweak = 'LC_ALL', 'LANG'
@@ -63,6 +104,76 @@ class XPath2ParserTest(test_xpath1_parser.XPath1ParserTest):
                 if self.current_env_vars[v] is not None:
                     os.environ[v] = self.current_env_vars[v]
 
+    def test_is_sequence_type_method(self):
+        self.assertTrue(self.parser.is_sequence_type('empty-sequence()'))
+        self.assertTrue(self.parser.is_sequence_type('xs:string'))
+        self.assertTrue(self.parser.is_sequence_type('xs:float+'))
+        self.assertTrue(self.parser.is_sequence_type('element()*'))
+        self.assertTrue(self.parser.is_sequence_type('item()?'))
+        self.assertTrue(self.parser.is_sequence_type('xs:untypedAtomic+'))
+        self.assertFalse(self.parser.is_sequence_type(10))
+        self.assertFalse(self.parser.is_sequence_type('empty-sequence()*'))
+        self.assertFalse(self.parser.is_sequence_type('unknown'))
+        self.assertFalse(self.parser.is_sequence_type('unknown?'))
+        self.assertFalse(self.parser.is_sequence_type('tns0:unknown'))
+
+    def test_match_sequence_type_method(self):
+        self.assertTrue(self.parser.match_sequence_type(None, 'empty-sequence()'))
+        self.assertTrue(self.parser.match_sequence_type([], 'empty-sequence()'))
+        self.assertFalse(self.parser.match_sequence_type('', 'empty-sequence()'))
+
+        self.assertFalse(self.parser.match_sequence_type('', 'empty-sequence()'))
+
+        root = self.etree.XML('<root><e1/><e2/><e3/></root>')
+        self.assertTrue(self.parser.match_sequence_type(root, 'element()'))
+        self.assertTrue(self.parser.match_sequence_type([root], 'element()'))
+        self.assertTrue(self.parser.match_sequence_type(root, 'element()', '?'))
+        self.assertTrue(self.parser.match_sequence_type(root, 'element()', '+'))
+        self.assertTrue(self.parser.match_sequence_type(root, 'element()', '*'))
+        self.assertFalse(self.parser.match_sequence_type(root[:], 'element()'))
+        self.assertFalse(self.parser.match_sequence_type(root[:], 'element()', '?'))
+        self.assertTrue(self.parser.match_sequence_type(root[:], 'element()', '+'))
+        self.assertTrue(self.parser.match_sequence_type(root[:], 'element()', '*'))
+
+        self.assertTrue(self.parser.match_sequence_type(UntypedAtomic(1), 'xs:untypedAtomic'))
+        self.assertFalse(self.parser.match_sequence_type(1, 'xs:untypedAtomic'))
+
+        self.assertTrue(self.parser.match_sequence_type('1', 'xs:string'))
+        self.assertFalse(self.parser.match_sequence_type(1, 'xs:string'))
+        self.assertFalse(self.parser.match_sequence_type('1', 'xs:unknown'))
+        self.assertFalse(self.parser.match_sequence_type('1', 'tns0:string'))
+
+    def test_check_variables_method(self):
+        self.parser.variable_types.update(
+            (k, get_sequence_type(v)) for k, v in self.variables.items()
+        )
+        self.assertEqual(self.parser.variable_types,
+                         {'values': 'xs:decimal+', 'myaddress': 'xs:string', 'word': 'xs:string'})
+
+        self.assertIsNone(self.parser.check_variables(
+            {'values': 1, 'myaddress': 'info@example.com', 'word': ''}
+        ))
+
+        with self.assertRaises(NameError) as ctx:
+            self.parser.check_variables({'values': 1})
+        self.assertIn("[err:XPST0008] Missing variable", str(ctx.exception))
+
+        with self.assertRaises(TypeError) as ctx:
+            self.parser.check_variables(
+                {'values': 1.0, 'myaddress': 'info@example.com', 'word': ''}
+            )
+        self.assertEqual("[err:XPDY0050] Unmatched sequence type for variable 'values'",
+                         str(ctx.exception))
+
+        with self.assertRaises(TypeError) as ctx:
+            self.parser.check_variables(
+                {'values': 1, 'myaddress': 'info@example.com', 'word': True}
+            )
+        self.assertEqual("[err:XPDY0050] Unmatched sequence type for variable 'word'",
+                         str(ctx.exception))
+
+        self.parser.variable_types.clear()
+
     def test_xpath_tokenizer(self):
         super(XPath2ParserTest, self).test_xpath_tokenizer()
         self.check_tokenizer("(: this is a comment :)",
@@ -84,7 +195,11 @@ class XPath2ParserTest(test_xpath1_parser.XPath1ParserTest):
 
     def test_xpath_comments(self):
         self.wrong_syntax("(: this is a comment :)")
+        self.check_value("(: this is a comment :) true()", True)
+        self.check_value("(: comment 1 :)(: comment 2 :) true()", True)
+        self.check_value("(: comment 1 :) true() (: comment 2 :)", True)
         self.wrong_syntax("(: this is a (: nested :) comment :)")
+        self.check_value("(: this is a (: nested :) comment :) true()", True)
         self.check_tree('child (: nasty (:nested :) axis comment :) ::B1', '(child (B1))')
         self.check_tree('child (: nasty "(: but not nested :)" axis comment :) ::B1',
                         '(child (B1))')
@@ -128,7 +243,7 @@ class XPath2ParserTest(test_xpath1_parser.XPath1ParserTest):
         root = self.etree.XML('<part discounted="false"><wholesale/><retail/></part>')
         self.check_selector(
             'if ($part/@discounted) then $part/wholesale else $part/retail',
-            root, [root[0]], variables={'part': root}
+            root, [root[0]], variables={'part': root}, variable_types={'part': 'element()'}
         )
         root = self.etree.XML('<widgets>'
                               '  <widget><unit-cost>25</unit-cost></widget>'
@@ -166,14 +281,20 @@ class XPath2ParserTest(test_xpath1_parser.XPath1ParserTest):
         self.check_value("every $x in (1, 2, 3), $y in (2, 3, 4) satisfies $x + $y = 4",
                          False, context)
 
+        self.check_value("some $x in (1, 2, 3), $y in (2, 3, 4) satisfies $x + $y = 7",
+                         True, context)
+        self.check_value("some $x in (1, 2, 3), $y in (2, 3, 4) satisfies $x + $y = 8",
+                         False, context)
+
         self.check_value('some $x in (1, 2, "cat") satisfies $x * 2 = 4', True, context)
         self.check_value('every $x in (1, 2, "cat") satisfies $x * 2 = 4', False, context)
 
     def test_for_expressions(self):
         # Cases from XPath 2.0 examples
         context = XPathContext(root=self.etree.XML('<dummy/>'))
-        self.check_value("for $i in (10, 20), $j in (1, 2) return ($i + $j)",
-                         [11, 12, 21, 22], context)
+        path = "for $i in (10, 20), $j in (1, 2) return ($i + $j)"
+        self.check_value(path, [11, 12, 21, 22], context)
+        self.check_source(path, path)
 
         root = self.etree.XML(
             """
@@ -231,8 +352,8 @@ class XPath2ParserTest(test_xpath1_parser.XPath1ParserTest):
         self.check_value("5 idiv 2", 2)
         self.check_value("-3.5 idiv -2", 1)
         self.check_value("-3.5 idiv 2", -1)
-        self.wrong_value("-3.5 idiv 0")
-        self.wrong_value("xs:float('INF') idiv 2")
+        self.check_value("-3.5 idiv 0", ZeroDivisionError)
+        self.check_value("xs:float('INF') idiv 2", OverflowError)
 
     def test_comparison_operators(self):
         super(XPath2ParserTest, self).test_comparison_operators()
@@ -259,26 +380,25 @@ class XPath2ParserTest(test_xpath1_parser.XPath1ParserTest):
 
     def test_boolean_evaluation_in_selector(self):
         context = XPathContext(self.etree.XML("""
-<collection>
-    <book>
-        <available>true</available>
-        <price>10.0</price>
-    </book>
-    <book>
-        <available>1</available>
-        <price>10.0</price>
-    </book>
-    <book>
-        <available>false</available>
-        <price>5.0</price>
-    </book>
-    <book>
-        <available>0</available>
-        <price>5.0</price>
-    </book>
-</collection>
-        """))
-        expr = "sum(//price)"
+        <collection>
+            <book>
+                <available>true</available>
+                <price>10.0</price>
+            </book>
+            <book>
+                <available>1</available>
+                <price>10.0</price>
+            </book>
+            <book>
+                <available>false</available>
+                <price>5.0</price>
+            </book>
+            <book>
+                <available>0</available>
+                <price>5.0</price>
+            </book>
+        </collection>"""))
+
         self.check_value("sum(//price)", 30, context)
         self.check_value("sum(//price[../available = 'true'])", 10, context)
         self.check_value("sum(//price[../available = 'false'])", 5, context)
@@ -291,7 +411,7 @@ class XPath2ParserTest(test_xpath1_parser.XPath1ParserTest):
         super(XPath2ParserTest, self).test_comparison_of_sequences()
 
         self.parser.compatibility_mode = True
-        self.check_value("(false(), false()) = 1", False)
+        self.wrong_type("(false(), false()) = 1")
         self.check_value("(false(), false()) = (false(), false())", True)
         self.check_value("(false(), false()) = (false(), false(), false())", True)
         self.check_value("(false(), false()) = (false(), true())", True)
@@ -357,7 +477,7 @@ class XPath2ParserTest(test_xpath1_parser.XPath1ParserTest):
 
                 with self.assertRaises(TypeError) as err:
                     self.check_value('. le 10', context=XPathContext(root))
-                self.assertIn('FORG0006', str(err.exception))  # Dynamic context error
+                self.assertIn('XPTY0004', str(err.exception))  # Dynamic context error
 
             schema = xmlschema.XMLSchema("""
                 <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
@@ -376,6 +496,15 @@ class XPath2ParserTest(test_xpath1_parser.XPath1ParserTest):
         self.check_selector("a[3] = (1 to 10, 30)", root, True)
         self.check_selector("a[4] = (1 to 10, 30)", root, False)
 
+    def test_unknown_axis(self):
+        self.wrong_syntax('unknown::node()', 'XPST0003')
+        self.wrong_syntax('A/unknown::node()', 'XPST0003')
+
+        self.parser.compatibility_mode = True
+        self.wrong_name('unknown::node()', 'XPST0010')
+        self.wrong_name('A/unknown::node()', 'XPST0010')
+        self.parser.compatibility_mode = False
+
     def test_predicate(self):
         super(XPath2ParserTest, self).test_predicate()
         root = self.etree.XML('<A><B1><B2/><B2/></B1><C1><C2/><C2/></C1></A>')
@@ -498,7 +627,7 @@ class XPath2ParserTest(test_xpath1_parser.XPath1ParserTest):
                          DayTimeDuration.fromstring('PT17H40M7S'))
         self.check_value(
             'xs:dayTimeDuration("P2DT53M11S") div xs:dayTimeDuration("P1DT10H")',
-            Decimal('1.437834967320261437908496732')
+            Decimal('1.43783496732026143790849673203')
         )
 
     def test_document_node_accessor(self):
@@ -647,8 +776,11 @@ class XPath2ParserTest(test_xpath1_parser.XPath1ParserTest):
         self.check_value(". instance of empty-sequence()", expected=False, context=context)
 
     def test_item_sequence_type(self):
-        self.check_value("4 treat as item()", [4])
-        self.check_value("() treat as item()", TypeError)
+        self.check_value("4 treat as item()", MissingContextError)
+
+        context = XPathContext(self.etree.XML('<root/>'))
+        self.check_value("4 treat as item()", [4], context)
+        self.check_value("() treat as item()", TypeError, context)
         self.wrong_syntax("item()")
 
         context = XPathContext(root=self.etree.XML('<A/>'))
@@ -660,9 +792,10 @@ class XPath2ParserTest(test_xpath1_parser.XPath1ParserTest):
         self.check_value("() instance of item()", expected=False, context=context)
 
     def test_static_analysis_phase(self):
-        self.check_value('fn:concat($word, fn:lower-case(" BETA"))', 'alpha beta')
-        self.check_value('fn:concat($word, fn:lower-case(10))', TypeError)
-        self.check_value('fn:concat($unknown, fn:lower-case(10))', TypeError)
+        context = XPathContext(self.etree.XML('<root/>'), variables=self.variables)
+        self.check_value('fn:concat($word, fn:lower-case(" BETA"))', 'alpha beta', context)
+        self.check_value('fn:concat($word, fn:lower-case(10))', TypeError, context)
+        self.check_value('fn:concat($unknown, fn:lower-case(10))', NameError, context)
 
     def test_instance_of_expression(self):
         element = self.etree.Element('schema')
@@ -728,7 +861,7 @@ class XPath2ParserTest(test_xpath1_parser.XPath1ParserTest):
     def test_logical_expressions_(self):
         super(XPath2ParserTest, self).test_logical_expressions()
 
-        if xmlschema is not None:
+        if xmlschema is not None and xmlschema.__version__ >= '1.2.3':
             schema = xmlschema.XMLSchema("""
                 <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
                   <xs:element name="root">
@@ -743,15 +876,15 @@ class XPath2ParserTest(test_xpath1_parser.XPath1ParserTest):
             with self.schema_bound_parser(schema.elements['root'].xpath_proxy):
                 root_token = self.parser.parse("(@a and not(@b)) or (not(@a) and @b)")
                 context = XPathContext(self.etree.XML('<root a="10" b="0"/>'))
-                self.assertTrue(root_token.evaluate(context=context) is True)
+                self.assertTrue(root_token.evaluate(context=context) is False)
                 context = XPathContext(self.etree.XML('<root a="10" b="1"/>'))
                 self.assertTrue(root_token.evaluate(context=context) is False)
                 context = XPathContext(self.etree.XML('<root a="10"/>'))
                 self.assertTrue(root_token.evaluate(context=context) is True)
                 context = XPathContext(self.etree.XML('<root a="0" b="10"/>'))
-                self.assertTrue(root_token.evaluate(context=context) is True)
-                context = XPathContext(self.etree.XML('<root b="0"/>'))
                 self.assertTrue(root_token.evaluate(context=context) is False)
+                context = XPathContext(self.etree.XML('<root b="0"/>'))
+                self.assertTrue(root_token.evaluate(context=context) is True)
 
     def test_element_decimal_cast(self):
         root = self.etree.XML('''
@@ -779,7 +912,79 @@ class XPath2ParserTest(test_xpath1_parser.XPath1ParserTest):
     def test_token_ambiguity(self):
         # Related to issue #27
         self.check_tokenizer("/is", ['/', 'is'])
-        self.check_value('/is', [])
+        context = XPathContext(self.etree.XML('<root/>'))
+        self.check_value('/is', [], context)
+        context = XPathContext(self.etree.XML('<is/>'))
+        self.check_value('/is', [context.root], context)
+
+    @unittest.skipIf(xmlschema is None, "xmlschema library required.")
+    def test_get_atomic_value(self):
+        schema = xmlschema.XMLSchema("""
+            <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
+              <xs:element name="a" type="aType"/>
+              <xs:complexType name="aType">
+                <xs:sequence>
+                  <xs:element name="b1" type="xs:int"/>
+                  <xs:element name="b2" type="xs:boolean"/>
+                </xs:sequence>
+              </xs:complexType>
+
+              <xs:element name="b" type="xs:int"/>
+              <xs:element name="c"/>
+
+              <xs:element name="d" type="dType"/>
+              <xs:simpleType name="dType">
+                <xs:restriction base="xs:float"/>
+              </xs:simpleType>
+
+              <xs:element name="e" type="eType"/>
+              <xs:simpleType name="eType">
+                <xs:union memberTypes="xs:string xs:integer xs:boolean"/>
+              </xs:simpleType>
+            </xs:schema>""")
+
+        token = self.parser.parse('true()')
+
+        self.assertEqual(self.parser.get_atomic_value(schema.elements['d'].type),
+                         UntypedAtomic('1'))
+
+        self.parser.schema = xmlschema.xpath.XMLSchemaProxy()
+        try:
+            token.string_value(schema.elements['a'].type)
+        finally:
+            self.parser.schema = None
+
+        self.parser.schema = xmlschema.xpath.XMLSchemaProxy(schema)
+        try:
+            with self.assertRaises(AttributeError) as err:
+                self.parser.get_atomic_value(schema)
+
+            value = self.parser.get_atomic_value(schema.elements['a'].type)
+            self.assertIsInstance(value, UntypedAtomic)
+            self.assertEqual(value, UntypedAtomic(value='1'))
+
+            value = self.parser.get_atomic_value(schema.elements['b'].type)
+            if not isinstance(value, int):
+                import pdb
+                pdb.set_trace()
+                self.parser.get_atomic_value(schema.elements['b'].type)
+
+            self.assertIsInstance(value, int)
+            self.assertEqual(value, 1)
+
+            value = self.parser.get_atomic_value(schema.elements['c'].type)
+            self.assertIsInstance(value, UntypedAtomic)
+            self.assertEqual(value, UntypedAtomic(value='1'))
+
+            value = self.parser.get_atomic_value(schema.elements['d'].type)
+            self.assertIsInstance(value, float)
+            self.assertEqual(value, 1.0)
+
+            value = self.parser.get_atomic_value(schema.elements['e'].type)
+            self.assertIsInstance(value, UntypedAtomic)
+            self.assertEqual(value, UntypedAtomic(value='1'))
+        finally:
+            self.parser.schema = None
 
 
 @unittest.skipIf(lxml_etree is None, "The lxml library is not installed")
