@@ -9,6 +9,7 @@
 # @author Davide Brunato <brunato@sissa.it>
 #
 import unittest
+from unittest.mock import patch
 import io
 import math
 import xml.etree.ElementTree as ElementTree
@@ -22,13 +23,29 @@ else:
     xmlschema.XMLSchema.meta_schema.build()
 
 from elementpath.exceptions import MissingContextError
-from elementpath.datatypes import UntypedAtomic
-from elementpath.namespaces import XSD_NAMESPACE
-from elementpath.xpath_nodes import AttributeNode, TypedAttribute, TypedElement, NamespaceNode
+from elementpath.datatypes import UntypedAtomic, QName
+from elementpath.namespaces import XSD_NAMESPACE, XPATH_FUNCTIONS_NAMESPACE
+from elementpath.xpath_nodes import AttributeNode, TypedAttribute, \
+    TypedElement, NamespaceNode, TextNode
 from elementpath.xpath_token import ordinal
 from elementpath.xpath_context import XPathContext
 from elementpath.xpath1_parser import XPath1Parser
 from elementpath.xpath2_parser import XPath2Parser
+from elementpath.schema_proxy import AbstractXsdType
+
+
+DummyXsdType = type(
+    'XsdType', (AbstractXsdType,),
+    dict(name=None, local_name=None, is_matching=lambda x: False, **{
+        k: lambda x: None for k in AbstractXsdType.__dict__ if k[0] != '_'
+    }))
+
+
+class Tagged(object):
+    tag = 'root'
+
+    def __repr__(self):
+        return 'Tagged(tag=%r)' % self.tag
 
 
 class XPath1TokenTest(unittest.TestCase):
@@ -46,93 +63,6 @@ class XPath1TokenTest(unittest.TestCase):
         self.assertEqual(ordinal(23), '23rd')
         self.assertEqual(ordinal(34), '34th')
 
-    def test_get_argument_method(self):
-        token = self.parser.symbol_table['true'](self.parser)
-
-        self.assertIsNone(token.get_argument(2))
-        with self.assertRaises(TypeError):
-            token.get_argument(1, required=True)
-
-    def test_select_results(self):
-        token = self.parser.parse('.')
-        elem = ElementTree.Element('A', attrib={'max': '30'})
-        elem.text = '10'
-
-        context = XPathContext(elem)
-        self.assertListEqual(list(token.select_results(context)), [elem])
-
-        context = XPathContext(elem, item=TypedElement(elem, 10))
-        self.assertListEqual(list(token.select_results(context)), [elem])
-
-        context = XPathContext(elem, item=AttributeNode('max', '30'))
-        self.assertListEqual(list(token.select_results(context)), ['30'])
-
-        context = XPathContext(elem, item=TypedAttribute(AttributeNode('max', '30'), 30))
-        self.assertListEqual(list(token.select_results(context)), [30])
-
-        attribute = namedtuple('XsdAttribute', 'name type')('max', 'xs:string')
-        context = XPathContext(elem, item=TypedAttribute(AttributeNode('max', attribute), 30))
-        self.assertListEqual(list(token.select_results(context)), [attribute])
-
-        context = XPathContext(elem, item=10)
-        self.assertListEqual(list(token.select_results(context)), [10])
-
-        context = XPathContext(elem, item='10')
-        self.assertListEqual(list(token.select_results(context)), ['10'])
-
-    def test_boolean_value_function(self):
-        token = self.parser.parse('true()')
-        elem = ElementTree.Element('A')
-        with self.assertRaises(TypeError):
-            token.boolean_value(elem)
-
-        self.assertFalse(token.boolean_value([]))
-        self.assertTrue(token.boolean_value([elem]))
-        self.assertFalse(token.boolean_value([0]))
-        self.assertTrue(token.boolean_value([1]))
-        with self.assertRaises(TypeError):
-            token.boolean_value([1, 1])
-        with self.assertRaises(TypeError):
-            token.boolean_value(elem)
-        self.assertFalse(token.boolean_value(0))
-        self.assertTrue(token.boolean_value(1))
-
-    def test_data_value_function(self):
-        token = self.parser.parse('true()')
-        self.assertIsNone(token.data_value(None))
-
-    def test_string_value_function(self):
-        token = self.parser.parse('true()')
-
-        document = ElementTree.parse(io.StringIO(u'<A>123<B1>456</B1><B2>789</B2></A>'))
-        element = ElementTree.Element('schema')
-        attribute = AttributeNode('id', '0212349350')
-        namespace = NamespaceNode('xs', 'http://www.w3.org/2001/XMLSchema')
-        comment = ElementTree.Comment('nothing important')
-        pi = ElementTree.ProcessingInstruction('action', 'nothing to do')
-        text = 'betelgeuse'
-        self.assertEqual(token.string_value(document), '123456789')
-        self.assertEqual(token.string_value(element), '')
-        self.assertEqual(token.string_value(attribute), '0212349350')
-        self.assertEqual(token.string_value(namespace), 'http://www.w3.org/2001/XMLSchema')
-        self.assertEqual(token.string_value(comment), 'nothing important')
-        self.assertEqual(token.string_value(pi), 'action nothing to do')
-        self.assertEqual(token.string_value(text), 'betelgeuse')
-        self.assertEqual(token.string_value(None), '')
-        self.assertEqual(token.string_value(10), '10')
-
-    def test_number_value_function(self):
-        token = self.parser.parse('true()')
-        self.assertEqual(token.number_value("19"), 19)
-        self.assertTrue(math.isnan(token.number_value("not a number")))
-
-    def test_compare_operator(self):
-        token1 = self.parser.parse('true()')
-        token2 = self.parser.parse('false()')
-        self.assertEqual(token1, token1)
-        self.assertNotEqual(token1, token2)
-        self.assertNotEqual(token2, 'false()')
-
     def test_arity_property(self):
         token = self.parser.parse('true()')
         self.assertEqual(token.symbol, 'true')
@@ -160,72 +90,310 @@ class XPath1TokenTest(unittest.TestCase):
         items = [tk for tk in token.iter()]
         self.assertListEqual(items, [token[0], token, token[1]])
 
+        token = self.parser.parse('/A/B[C]/D/@a')
+        self.assertEqual(token.tree, '(/ (/ (/ (/ (A)) ([ (B) (C))) (D)) (@ (a)))')
+        self.assertListEqual(list(tk.value for tk in token.iter()),
+                             ['/', 'A', '/', 'B', '[', 'C', '/', 'D', '/', '@', 'a'])
+        self.assertListEqual(list(tk.value for tk in token.iter('(name)')),
+                             ['A', 'B', 'C', 'D', 'a'])
+
+        self.assertListEqual(list(tk.source for tk in token.iter('/')),
+                             ['/ A', '/ A / B[C]', '/ A / B[C] / D', '/ A / B[C] / D / @ a'])
+
+    def test_iter_leaf_elements_method(self):
+        token = self.parser.parse('2 + 5')
+        self.assertListEqual(list(token.iter_leaf_elements()), [])
+
+        token = self.parser.parse('/A/B[C]/D/@a')
+        self.assertListEqual(list(token.iter_leaf_elements()), [])
+
+        token = self.parser.parse('/A/B[C]/D')
+        self.assertListEqual(list(token.iter_leaf_elements()), ['D'])
+
+        token = self.parser.parse('/A/B[C]')
+        self.assertEqual(token.tree, '(/ (/ (A)) ([ (B) (C)))')
+        self.assertListEqual(list(token.iter_leaf_elements()), ['B'])
+
+    def test_get_argument_method(self):
+        token = self.parser.symbol_table['true'](self.parser)
+
+        self.assertIsNone(token.get_argument(2))
+        with self.assertRaises(TypeError):
+            token.get_argument(1, required=True)
+
+    @patch.multiple(DummyXsdType, is_simple=lambda x: False, has_simple_content=lambda x: True)
+    def test_select_results(self):
+        token = self.parser.parse('.')
+        elem = ElementTree.Element('A', attrib={'max': '30'})
+        elem.text = '10'
+        xsd_type = DummyXsdType()
+
+        context = XPathContext(elem)
+        self.assertListEqual(list(token.select_results(context)), [elem])
+
+        context = XPathContext(elem, item=TypedElement(elem, xsd_type, 10))
+        self.assertListEqual(list(token.select_results(context)), [elem])
+
+        context = XPathContext(elem, item=AttributeNode('max', '30'))
+        self.assertListEqual(list(token.select_results(context)), ['30'])
+
+        item = TypedAttribute(AttributeNode('max', '30'), xsd_type, 30)
+        context = XPathContext(elem, item=item)
+        self.assertListEqual(list(token.select_results(context)), [30])
+
+        attribute = namedtuple('XsdAttribute', 'name local_name type')('max', 'max', xsd_type)
+        item = TypedAttribute(AttributeNode('max', attribute), xsd_type, 30)
+        context = XPathContext(elem, item=item)
+        self.assertListEqual(list(token.select_results(context)), [attribute])
+
+        context = XPathContext(elem, item=10)
+        self.assertListEqual(list(token.select_results(context)), [10])
+
+        context = XPathContext(elem, item='10')
+        self.assertListEqual(list(token.select_results(context)), ['10'])
+
+    def test_cast_to_double(self):
+        token = self.parser.parse('.')
+        self.assertEqual(token.cast_to_double(1), 1.0)
+
+        with self.assertRaises(ValueError) as ctx:
+            token.cast_to_double('nan')
+        self.assertIn('FORG0001', str(ctx.exception))
+
+        if self.parser.version != '1.0':
+            self.parser._xsd_version = '1.1'
+            self.assertEqual(token.cast_to_double('1'), 1.0)
+            self.parser._xsd_version = '1.0'
+
+    def test_boolean_value_function(self):
+        token = self.parser.parse('true()')
+        elem = ElementTree.Element('A')
+        with self.assertRaises(TypeError):
+            token.boolean_value(elem)
+
+        self.assertFalse(token.boolean_value([]))
+        self.assertTrue(token.boolean_value([elem]))
+        self.assertFalse(token.boolean_value([0]))
+        self.assertTrue(token.boolean_value([1]))
+        with self.assertRaises(TypeError):
+            token.boolean_value([1, 1])
+        with self.assertRaises(TypeError):
+            token.boolean_value(elem)
+        self.assertFalse(token.boolean_value(0))
+        self.assertTrue(token.boolean_value(1))
+        self.assertTrue(token.boolean_value(1.0))
+        self.assertFalse(token.boolean_value(None))
+
+    def test_data_value_function(self):
+        token = self.parser.parse('true()')
+
+        if self.parser.version != '1.0':
+            with patch.multiple(DummyXsdType(), is_simple=lambda x: False,
+                                has_simple_content=lambda x: True) as xsd_type:
+                obj = TypedElement(ElementTree.XML('<age>19</age>'), xsd_type, 19)
+                self.assertEqual(token.data_value(obj), 19)
+
+        obj = AttributeNode('age', '19')
+        self.assertEqual(token.data_value(obj), UntypedAtomic('19'))
+
+        obj = NamespaceNode('tns', 'http://xpath.test/ns')
+        self.assertEqual(token.data_value(obj), 'http://xpath.test/ns')
+
+        obj = TextNode('19')
+        self.assertEqual(token.data_value(obj), UntypedAtomic('19'))
+
+        obj = ElementTree.XML('<root>a<e1>b</e1>c<e2>d</e2>e</root>')
+        self.assertEqual(token.data_value(obj), UntypedAtomic('abcde'))
+
+        obj = ElementTree.parse(io.StringIO('<root>a<e1>b</e1>c<e2>d</e2>e</root>'))
+        self.assertEqual(token.data_value(obj), UntypedAtomic('abcde'))
+
+        obj = ElementTree.Comment("foo bar")
+        self.assertEqual(token.data_value(obj), 'foo bar')
+
+        obj = ElementTree.ProcessingInstruction('action', 'nothing to do')
+        self.assertEqual(token.data_value(obj), 'action nothing to do')
+
+        self.assertIsNone(token.data_value(None))
+        self.assertEqual(token.data_value(19), 19)
+        self.assertEqual(token.data_value('19'), '19')
+        self.assertFalse(token.data_value(False))
+
+        tagged_object = Tagged()
+        self.assertIsNone(token.data_value(tagged_object))
+
+    def test_string_value_function(self):
+        token = self.parser.parse('true()')
+
+        document = ElementTree.parse(io.StringIO(u'<A>123<B1>456</B1><B2>789</B2></A>'))
+        element = ElementTree.Element('schema')
+        attribute = AttributeNode('id', '0212349350')
+        namespace = NamespaceNode('xs', 'http://www.w3.org/2001/XMLSchema')
+        comment = ElementTree.Comment('nothing important')
+        pi = ElementTree.ProcessingInstruction('action', 'nothing to do')
+        text = 'betelgeuse'
+        self.assertEqual(token.string_value(document), '123456789')
+        self.assertEqual(token.string_value(element), '')
+        self.assertEqual(token.string_value(attribute), '0212349350')
+        self.assertEqual(token.string_value(namespace), 'http://www.w3.org/2001/XMLSchema')
+        self.assertEqual(token.string_value(comment), 'nothing important')
+        self.assertEqual(token.string_value(pi), 'action nothing to do')
+        self.assertEqual(token.string_value(text), 'betelgeuse')
+        self.assertEqual(token.string_value(None), '')
+        self.assertEqual(token.string_value(10), '10')
+        self.assertEqual(token.string_value(1e99), '1E99')
+        self.assertEqual(token.string_value(float('nan')), 'NaN')
+        self.assertEqual(token.string_value(float('inf')), 'INF')
+        self.assertEqual(token.string_value(float('-inf')), '-INF')
+
+        self.assertEqual(token.string_value(()), '()')
+
+        tagged_object = Tagged()
+        self.assertEqual(token.string_value(tagged_object), "Tagged(tag='root')")
+
+        with patch.multiple(DummyXsdType, is_simple=lambda x: True):
+            xsd_type = DummyXsdType()
+            typed_elem = TypedElement(elem=element, type=xsd_type, value=10)
+            self.assertEqual(token.string_value(typed_elem), '10')
+            typed_elem = TypedElement(elem=element, type=xsd_type, value=None)
+            self.assertEqual(token.string_value(typed_elem), '')
+
+    def test_number_value_function(self):
+        token = self.parser.parse('true()')
+        self.assertEqual(token.number_value("19"), 19)
+        self.assertTrue(math.isnan(token.number_value("not a number")))
+
+    def test_compare_operator(self):
+        token1 = self.parser.parse('true()')
+        token2 = self.parser.parse('false()')
+        self.assertEqual(token1, token1)
+        self.assertNotEqual(token1, token2)
+        self.assertNotEqual(token2, 'false()')
+
+    def test_expected_method(self):
+        token = self.parser.parse('.')
+        self.assertIsNone(token.expected('.'))
+
+        with self.assertRaises(SyntaxError) as ctx:
+            raise token.expected('*')
+        self.assertIn('XPST0003', str(ctx.exception))
+
+    def test_unexpected_method(self):
+        token = self.parser.parse('.')
+        self.assertIsNone(token.unexpected('*'))
+
+        with self.assertRaises(SyntaxError) as ctx:
+            raise token.unexpected('.')
+        self.assertIn('XPST0003', str(ctx.exception))
+
+        with self.assertRaises(SyntaxError) as ctx:
+            raise token.unexpected('.', message="unknown error")
+        self.assertIn('XPST0003', str(ctx.exception))
+        self.assertIn('unknown error', str(ctx.exception))
+
+        with self.assertRaises(TypeError) as ctx:
+            raise token.unexpected('.', code='XPST0017')
+        self.assertIn('XPST0017', str(ctx.exception))
+
+    def test_xpath_error_code(self):
+        parser = XPath2Parser()
+        token = parser.parse('.')
+
+        self.assertEqual(token.error_code('XPST0003'), 'err:XPST0003')
+        parser.namespaces['error'] = parser.namespaces.pop('err')
+        self.assertEqual(token.error_code('XPST0003'), 'error:XPST0003')
+        parser.namespaces.pop('error')
+        self.assertEqual(token.error_code('XPST0003'), 'XPST0003')
+
+    def test_xpath_error(self):
+        token = self.parser.parse('.')
+
+        with self.assertRaises(ValueError) as ctx:
+            raise token.error('xml:XPST0003')
+        self.assertIn('XPTY0004', str(ctx.exception))
+        self.assertIn("required for error code", str(ctx.exception))
+
+        with self.assertRaises(ValueError) as ctx:
+            raise token.error('err:err:XPST0003')
+        self.assertIn('XPTY0004', str(ctx.exception))
+        self.assertIn("is not a QName", str(ctx.exception))
+
+        with self.assertRaises(ValueError) as ctx:
+            raise token.error('XPST9999')
+        self.assertIn('XPTY0004', str(ctx.exception))
+        self.assertIn("unknown XPath error code", str(ctx.exception))
+
     def test_xpath_error_shortcuts(self):
         token = self.parser.parse('.')
 
-        with self.assertRaises(ValueError) as err:
-            token.wrong_value()
-        self.assertIn('FOCA0002', str(err.exception))
+        with self.assertRaises(ValueError) as ctx:
+            raise token.wrong_value()
+        self.assertIn('FOCA0002', str(ctx.exception))
 
-        with self.assertRaises(TypeError) as err:
-            token.wrong_type()
-        self.assertIn('FORG0006', str(err.exception))
+        with self.assertRaises(TypeError) as ctx:
+            raise token.wrong_type()
+        self.assertIn('FORG0006', str(ctx.exception))
 
-        with self.assertRaises(ValueError) as err:
-            token.missing_schema()
-        self.assertIn('XPST0001', str(err.exception))
+        with self.assertRaises(ValueError) as ctx:
+            raise token.missing_schema()
+        self.assertIn('XPST0001', str(ctx.exception))
 
-        with self.assertRaises(MissingContextError) as err:
-            token.missing_context()
-        self.assertIn('XPDY0002', str(err.exception))
+        with self.assertRaises(MissingContextError) as ctx:
+            raise token.missing_context()
+        self.assertIn('XPDY0002', str(ctx.exception))
 
-        with self.assertRaises(TypeError) as err:
-            token.wrong_context_type()
-        self.assertIn('XPTY0004', str(err.exception))
+        with self.assertRaises(TypeError) as ctx:
+            raise token.wrong_context_type()
+        self.assertIn('XPTY0004', str(ctx.exception))
 
-        with self.assertRaises(ValueError) as err:
-            token.missing_sequence()
-        self.assertIn('XPST0005', str(err.exception))
+        with self.assertRaises(ValueError) as ctx:
+            raise token.missing_sequence()
+        self.assertIn('XPST0005', str(ctx.exception))
 
-        with self.assertRaises(NameError) as err:
-            token.missing_name()
-        self.assertIn('XPST0008', str(err.exception))
+        with self.assertRaises(NameError) as ctx:
+            raise token.missing_name()
+        self.assertIn('XPST0008', str(ctx.exception))
 
-        with self.assertRaises(NameError) as err:
-            token.missing_axis()
-        self.assertIn('XPST0010', str(err.exception))
+        if self.parser.compatibility_mode:
+            with self.assertRaises(NameError) as ctx:
+                raise token.missing_axis()
+            self.assertIn('XPST0010', str(ctx.exception))
+        else:
+            with self.assertRaises(SyntaxError) as ctx:
+                raise token.missing_axis()
+            self.assertIn('XPST0003', str(ctx.exception))
 
-        with self.assertRaises(TypeError) as err:
-            token.wrong_nargs()
-        self.assertIn('XPST0017', str(err.exception))
+        with self.assertRaises(TypeError) as ctx:
+            raise token.wrong_nargs()
+        self.assertIn('XPST0017', str(ctx.exception))
 
-        with self.assertRaises(TypeError) as err:
-            token.wrong_step_result()
-        self.assertIn('XPTY0018', str(err.exception))
+        with self.assertRaises(TypeError) as ctx:
+            raise token.wrong_step_result()
+        self.assertIn('XPTY0018', str(ctx.exception))
 
-        with self.assertRaises(TypeError) as err:
-            token.wrong_intermediate_step_result()
-        self.assertIn('XPTY0019', str(err.exception))
+        with self.assertRaises(TypeError) as ctx:
+            raise token.wrong_intermediate_step_result()
+        self.assertIn('XPTY0019', str(ctx.exception))
 
-        with self.assertRaises(TypeError) as err:
-            token.wrong_axis_argument()
-        self.assertIn('XPTY0020', str(err.exception))
+        with self.assertRaises(TypeError) as ctx:
+            raise token.wrong_axis_argument()
+        self.assertIn('XPTY0020', str(ctx.exception))
 
-        with self.assertRaises(TypeError) as err:
-            token.wrong_sequence_type()
-        self.assertIn('XPDY0050', str(err.exception))
+        with self.assertRaises(TypeError) as ctx:
+            raise token.wrong_sequence_type()
+        self.assertIn('XPDY0050', str(ctx.exception))
 
-        with self.assertRaises(NameError) as err:
-            token.unknown_atomic_type()
-        self.assertIn('XPST0051', str(err.exception))
+        with self.assertRaises(NameError) as ctx:
+            raise token.unknown_atomic_type()
+        self.assertIn('XPST0051', str(ctx.exception))
 
-        with self.assertRaises(NameError) as err:
-            token.wrong_target_type()
-        self.assertIn('XPST0080', str(err.exception))
+        with self.assertRaises(NameError) as ctx:
+            raise token.wrong_target_type()
+        self.assertIn('XPST0080', str(ctx.exception))
 
-        with self.assertRaises(NameError) as err:
-            token.unknown_namespace()
-        self.assertIn('XPST0081', str(err.exception))
+        with self.assertRaises(NameError) as ctx:
+            raise token.unknown_namespace()
+        self.assertIn('XPST0081', str(ctx.exception))
 
 
 class XPath2TokenTest(XPath1TokenTest):
@@ -234,6 +402,31 @@ class XPath2TokenTest(XPath1TokenTest):
     def setUpClass(cls):
         cls.parser = XPath2Parser(namespaces={'xs': XSD_NAMESPACE, 'tst': "http://xpath.test/ns"})
 
+    def test_bind_namespace_method(self):
+        token = self.parser.parse('true()')
+        self.assertIsNone(token.bind_namespace(XPATH_FUNCTIONS_NAMESPACE))
+        with self.assertRaises(TypeError) as ctx:
+            token.bind_namespace(XSD_NAMESPACE)
+        self.assertIn('XPST0017', str(ctx.exception))
+        self.assertIn("a name, a wildcard or a constructor", str(ctx.exception))
+
+        token = self.parser.parse("xs:string(10.1)")
+        with self.assertRaises(TypeError) as ctx:
+            token.bind_namespace(XSD_NAMESPACE)
+        self.assertIn('XPST0017', str(ctx.exception))
+        self.assertIn("a name, a wildcard or a constructor", str(ctx.exception))
+
+        self.assertIsNone(token[1].bind_namespace(XSD_NAMESPACE))
+        with self.assertRaises(TypeError) as ctx:
+            token[1].bind_namespace(XPATH_FUNCTIONS_NAMESPACE)
+        self.assertIn("a function expected", str(ctx.exception))
+
+        token = self.parser.parse("tst:foo")
+        with self.assertRaises(SyntaxError) as ctx:
+            token.bind_namespace('http://xpath.test/ns')
+        self.assertIn('XPST0003', str(ctx.exception))
+        self.assertIn("a name, a wildcard, a function or a constructor", str(ctx.exception))
+
     @unittest.skipIf(xmlschema is None, "xmlschema library required.")
     def test_add_xsd_type(self):
         schema = xmlschema.XMLSchema("""
@@ -245,7 +438,7 @@ class XPath2TokenTest(XPath1TokenTest):
 
         root_token = self.parser.parse('a1')
         self.assertIsNone(root_token.xsd_types)
-        root_token.add_xsd_type('a1', schema.meta_schema.types['int'])
+        root_token.add_xsd_type(schema.elements['a1'])
         self.assertEqual(root_token.xsd_types, {'a1': schema.meta_schema.types['int']})
 
         self.parser.schema = xmlschema.xpath.XMLSchemaProxy(schema)
@@ -362,36 +555,44 @@ class XPath2TokenTest(XPath1TokenTest):
             root_token = self.parser.parse('root')
             self.assertEqual(root_token.xsd_types, {'root': schema.meta_schema.types['int']})
 
-            obj = root_token.match_xsd_type(schema.elements['root'], 'root')
-            self.assertIsInstance(obj, TypedElement)
+            context = XPathContext(root=schema)
+            obj = list(root_token.select_xsd_nodes(context, 'root'))
+            self.assertIsInstance(obj[0], TypedElement)
             self.assertEqual(root_token.xsd_types, {'root': schema.meta_schema.types['int']})
 
+            context.axis = 'self'
             root_token.xsd_types = None
-            root_token.match_xsd_type(schema, 'root')
+            list(root_token.select_xsd_nodes(context, 'root'))
             self.assertIsNone(root_token.xsd_types)
 
-            obj = root_token.match_xsd_type(schema.elements['root'], 'root')
-            self.assertIsInstance(obj, TypedElement)
+            context.axis = None
+            obj = list(root_token.select_xsd_nodes(context, 'root'))
+            self.assertIsInstance(obj[0], TypedElement)
 
-            obj = root_token.match_xsd_type(schema.meta_schema.types['string'], 'root')
-            self.assertIsNone(obj)
+            context = XPathContext(root=schema.meta_schema)
+            obj = list(root_token.select_xsd_nodes(context, 'root'))
+            self.assertListEqual(obj, [])
 
             root_token = self.parser.parse('@a')
             self.assertEqual(root_token[0].xsd_types, {'a': schema.meta_schema.types['string']})
 
             attribute = AttributeNode('a', schema.attributes['a'])
-            obj = root_token.match_xsd_type(attribute, 'a')
-            self.assertIsInstance(obj, TypedAttribute)
+            context = XPathContext(root=schema.meta_schema, item=attribute, axis='self')
+
+            obj = list(root_token.select_xsd_nodes(context, 'a'))
+            self.assertIsInstance(obj[0], TypedAttribute)
             self.assertEqual(root_token[0].xsd_types, {'a': schema.meta_schema.types['string']})
 
             root_token.xsd_types = None
-            root_token.match_xsd_type(schema, 'a')
+            context = XPathContext(root=schema)
+            list(root_token.select_xsd_nodes(context, 'a'))
             self.assertIsNone(root_token.xsd_types)
 
-            obj = root_token.match_xsd_type(attribute, 'a')
-            self.assertIsInstance(obj, TypedAttribute)
-            self.assertEqual(obj[0], attribute)
-            self.assertIsInstance(obj[1], str)
+            context = XPathContext(root=schema.meta_schema, item=attribute, axis='self')
+            obj = list(root_token.select_xsd_nodes(context, 'a'))
+            self.assertIsInstance(obj[0], TypedAttribute)
+            self.assertEqual(obj[0][0], attribute)
+            self.assertIsInstance(obj[0][-1], str)
             self.assertEqual(root_token[0].xsd_types, {'a': schema.meta_schema.types['string']})
 
         finally:
@@ -419,9 +620,16 @@ class XPath2TokenTest(XPath1TokenTest):
             self.assertEqual(xsd_type, schema.meta_schema.types['int'])
             self.assertIsNone(root_token.get_xsd_type('node'))
 
-            root_token.add_xsd_type('node', schema.meta_schema.types['float'])
-            root_token.add_xsd_type('node', schema.meta_schema.types['boolean'])
-            root_token.add_xsd_type('node', schema.meta_schema.types['decimal'])
+            TestElement = namedtuple('XsdElement', 'name local_name type')
+            root_token.add_xsd_type(
+                TestElement('node', 'node', schema.meta_schema.types['float'])
+            )
+            root_token.add_xsd_type(
+                TestElement('node', 'node', schema.meta_schema.types['boolean'])
+            )
+            root_token.add_xsd_type(
+                TestElement('node', 'node', schema.meta_schema.types['decimal'])
+            )
 
             xsd_type = root_token.get_xsd_type('node')
             self.assertEqual(xsd_type, schema.meta_schema.types['float'])
@@ -464,7 +672,8 @@ class XPath2TokenTest(XPath1TokenTest):
 
             self.assertEqual(root_token.get_xsd_type(elem), schema.types['aType'])
 
-            root_token.add_xsd_type('a', schema.meta_schema.types['float'])
+            TestElement = namedtuple('XsdElement', 'name local_name type')
+            root_token.add_xsd_type(TestElement('a', 'a', schema.meta_schema.types['float']))
             self.assertEqual(root_token.get_xsd_type(elem), schema.types['aType'])
 
             root_token.xsd_types['a'].insert(0, schema.meta_schema.types['boolean'])
@@ -492,7 +701,8 @@ class XPath2TokenTest(XPath1TokenTest):
             elem.text = '49'
             node = root_token.get_typed_node(elem)
             self.assertIsInstance(node, TypedElement)
-            self.assertEqual(node[1], 49)
+            self.assertIsInstance(node[1], xmlschema.XsdType)
+            self.assertEqual(node[2], 49)
 
             elem.text = 'beta'
             with self.assertRaises(TypeError) as err:
@@ -504,11 +714,13 @@ class XPath2TokenTest(XPath1TokenTest):
             elem.text = '36'
             node = root_token.get_typed_node(elem)
             self.assertIsInstance(node, TypedElement)
-            self.assertIsInstance(node[1], UntypedAtomic)
-            self.assertEqual(node[1], 36)
+            self.assertIsInstance(node[1], xmlschema.XsdType)
+            self.assertIsInstance(node[2], UntypedAtomic)
+            self.assertEqual(node[2], 36)
 
             root_token.xsd_types['root'] = schema.meta_schema.types['anyType']
-            self.assertIs(root_token.get_typed_node(elem), elem)
+            node = root_token.get_typed_node(elem)
+            self.assertIs(node[0], elem)
 
             root_token = self.parser.parse('@a')
             self.assertEqual(root_token[0].xsd_types, {'a': schema.meta_schema.types['int']})
@@ -516,13 +728,15 @@ class XPath2TokenTest(XPath1TokenTest):
             attribute = AttributeNode('a', '10')
             node = root_token[0].get_typed_node(attribute)
             self.assertIsInstance(node, TypedAttribute)
-            self.assertEqual(node[1], 10)
+            self.assertIsInstance(node[1], xmlschema.XsdType)
+            self.assertEqual(node[2], 10)
 
             root_token[0].xsd_types['a'] = schema.meta_schema.types['anyType']
             node = root_token[0].get_typed_node(attribute)
             self.assertIsInstance(node, TypedAttribute)
-            self.assertIsInstance(node[1], UntypedAtomic)
-            self.assertEqual(node[1], 10)
+            self.assertIsInstance(node[1], xmlschema.XsdType)
+            self.assertIsInstance(node[2], UntypedAtomic)
+            self.assertEqual(node[2], 10)
 
         finally:
             self.parser.schema = None
@@ -545,72 +759,6 @@ class XPath2TokenTest(XPath1TokenTest):
             finally:
                 self.parser.schema = None
 
-    @unittest.skipIf(xmlschema is None, "xmlschema library required.")
-    def test_schema_node_value(self):
-        schema = xmlschema.XMLSchema("""
-            <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
-              <xs:element name="a" type="aType"/>
-              <xs:complexType name="aType">
-                <xs:sequence>
-                  <xs:element name="b1" type="xs:int"/>
-                  <xs:element name="b2" type="xs:boolean"/>
-                </xs:sequence>
-              </xs:complexType>
-              
-              <xs:element name="b" type="xs:int"/>
-              <xs:element name="c"/>
-              
-              <xs:element name="d" type="dType"/>
-              <xs:simpleType name="dType">
-                <xs:restriction base="xs:float"/>
-              </xs:simpleType>
-              
-              <xs:element name="e" type="eType"/>
-              <xs:simpleType name="eType">
-                <xs:union memberTypes="xs:string xs:integer xs:boolean"/>
-              </xs:simpleType>
-            </xs:schema>""")
-
-        token = self.parser.parse('true()')
-
-        with self.assertRaises(ValueError) as err:
-            token.schema_node_value(schema.elements['d'])
-        self.assertIn('XPST0001', str(err.exception))
-
-        self.parser.schema = xmlschema.xpath.XMLSchemaProxy()
-        try:
-            token.string_value(schema.elements['a'])
-        finally:
-            self.parser.schema = None
-
-        self.parser.schema = xmlschema.xpath.XMLSchemaProxy(schema)
-        try:
-            with self.assertRaises(TypeError) as err:
-                token.schema_node_value(schema)
-            self.assertIn('FORG0006', str(err.exception))
-
-            value = token.schema_node_value(schema.elements['a'])
-            self.assertIsInstance(value, UntypedAtomic)
-            self.assertEqual(value, UntypedAtomic(value=''))
-
-            value = token.schema_node_value(schema.elements['b'])
-            self.assertIsInstance(value, int)
-            self.assertEqual(value, 1)
-
-            value = token.schema_node_value(schema.elements['c'])
-            self.assertIsInstance(value, UntypedAtomic)
-            self.assertEqual(value, UntypedAtomic(value='1'))
-
-            value = token.schema_node_value(schema.elements['d'])
-            self.assertIsInstance(value, float)
-            self.assertEqual(value, 1.0)
-
-            value = token.schema_node_value(schema.elements['e'])
-            self.assertIsInstance(value, UntypedAtomic)
-            self.assertEqual(value, UntypedAtomic(value='1'))
-        finally:
-            self.parser.schema = None
-
 
 if __name__ == '__main__':
     unittest.main()
